package com.devbeh.managers;


import java.io.File;
import java.util.Map;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Future;
import java.util.stream.Collectors;

import com.devbeh.ActionClass;
import com.devbeh.Actor;
import com.devbeh.DetectionClass;
import com.devbeh.FrameIdAndBytes;
import com.devbeh.MetricService;
import com.devbeh.data.Action;
import com.devbeh.data.DetectionEntryOrBuilder;
import com.devbeh.data.DistanceEstimationEntry;
import com.devbeh.data.PerceptionResponse;
import com.devbeh.data.PerceptionResponseOrBuilder;
import com.devbeh.data.Point3d;
import com.devbeh.data.PoseEstimationEntry;
import com.devbeh.data.TrackingEntryOrBuilder;
import org.bytedeco.javacv.FFmpegFrameRecorder;
import org.bytedeco.javacv.Frame;
import org.bytedeco.javacv.FrameRecorder;
import org.bytedeco.javacv.OpenCVFrameConverter;
import org.bytedeco.opencv.global.opencv_imgproc;
import org.bytedeco.opencv.opencv_core.Mat;
import org.bytedeco.opencv.opencv_core.Point;
import org.bytedeco.opencv.opencv_core.Rect;
import org.bytedeco.opencv.opencv_core.Scalar;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * Используется для формирования файла с отмеченными результатами машинного восприятия: обрамляющими рамками объектов,
 * скелетами людей и т. д. с целью дебага.
 */
public class PerceptionVisualizationManager extends Actor implements FrameConsumer, TrackingConsumer, PoseDataConsumer,
        DistanceDataConsumer, ActionsConsumer
{
    private static final Logger LOGGER = LoggerFactory.getLogger(PerceptionVisualizationManager.class);

    static final Scalar[] COLORS = new Scalar[] {
            new Scalar(255,0,0,1),
            new Scalar(0,255,0,1),
            new Scalar(0,0,255,1),
            new Scalar(255,255,0,1),
            new Scalar(0,255,255,1),
            new Scalar(255,0,255,1),
            new Scalar(192,192,192,1),
            new Scalar(128,128,128,1),
            new Scalar(128,0,0,1),
            new Scalar(128,128,0,1),
            new Scalar(0,128,0,1),
            new Scalar(128,0,128,1),
            new Scalar(0,128,128,1),
            new Scalar(0,0,128,1)};

    private static final int[] POSE_PARENT_IDS =
            new int[]{0, 0, 1, 2, 0, 4, 5, 0, 17, 17, 8, 17, 11, 12, 17, 14, 15, 0};

    /**
     * Осуществляет преобразование кадра из Mat формата.
     */
    private final OpenCVFrameConverter.ToMat matConverter = new OpenCVFrameConverter.ToMat();

    /**
     * Имя данного менеджера.
     */
    private final String managerName;

    /**
     * Файл для записи видеозаписи, содержащей результаты машинного восприятия - обрамляющие рамки объектов, классы,
     * трехмерные скелеты людей, расстояния от камеры и др.
     */
    private final File resultVideoFile;

    /**
     * Номер кадра, с которого необходимо формировать видео.
     */
    private final long initialConsumedFrameId;

    /**
     * Включена ли отправка видеозаписи.
     */
    private final boolean enabled;

    /**
     * Предоставляет fps обрабатываемого видеопотока. Используется для создания результирующего видео.
     */
    private final Future<Double> fpsFuture;

    /**
     * Используется для остановки конвейера обработки видео в случае критичной ошибки.
     */
    private final Runnable pipelineStopper;

    /**
     * Используется для отображения метрик на странице metrics.
     */
    private final MetricService metricService;

    /**
     * Очередь кадров для обработки.
     */
    private final BlockingQueue<FrameIdAndBytes> frameQueue;

    /**
     * Очередь результатов оценки отслеживания объектов.
     */
    private final BlockingQueue<PerceptionResponseOrBuilder> trackingQueue;

    /**
     * Очередь результатов оценки скелетов людей.
     */
    private final BlockingQueue<PerceptionResponse> poseQueue;

    /**
     * Очередь результатов оценки положения объектов в координатах камеры.
     */
    private final BlockingQueue<PerceptionResponse> distanceQueue;

    /**
     * Очередь результатов детектирования действий людей.
     */
    private final BlockingQueue<PerceptionResponseOrBuilder> actionsQueue;

    /**
     * Рекордер для формирования результирующей видеозаписи или null.
     */
    private FrameRecorder frameRecorder;

    /**
     * Текущий обрабатываемый кадр или null.
     */
    private FrameIdAndBytes currentFrameInfo;

    /**
     * Результат отслеживания для текущего обрабатываемого кадра или null.
     */
    private PerceptionResponseOrBuilder currentTrackingResult;

    /**
     * Результат оценки трехмерных поз людей для текущего обрабатываемого кадра или null.
     */
    private PerceptionResponse currentPoseData;

    /**
     * Результат оценки положения объектов в системе координат камеры для текущего обрабатываемого кадра или null.
     */
    private PerceptionResponse currentDistanceResult;

    /**
     * Результат детектирования действий людей для текущего обрабатываемого кадра.
     */
    private PerceptionResponseOrBuilder currentActionsResult;

    /**
     * Имя метрики последнего обработанного идентификатора кадра.
     */
    private final String lastFrameIdMetricName;


    public PerceptionVisualizationManager(String videoName, String resultVideoDirPath, long initialConsumedFrameId,
            boolean enabled, int maxQueuedRequests, Future<Double> fpsFuture, Runnable pipelineStopper,
            MetricService metricService)
    {
        this.managerName = "PerceptionVisualizationManager-" + videoName;
        this.resultVideoFile = new File(resultVideoDirPath, videoName + ".mp4");
        this.initialConsumedFrameId = initialConsumedFrameId;
        this.enabled = enabled;
        this.fpsFuture = fpsFuture;
        this.pipelineStopper = pipelineStopper;
        this.metricService = metricService;
        this.frameQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.trackingQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.poseQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.distanceQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.actionsQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.lastFrameIdMetricName = videoName + ".pvm.lastFrameId";
    }


    private void submitLastFrameId(long frameId)
    {
        metricService.submit(lastFrameIdMetricName, frameId);
    }


    @Override
    public long getInitialLastConsumedFrameId() throws InterruptedException
    {
        return enabled ? initialConsumedFrameId : Long.MAX_VALUE;
    }


    @Override
    public void putFrame(FrameIdAndBytes frameIdAndBytes) throws InterruptedException
    {
        if (enabled && !getDoStop())
        {
            frameQueue.put(frameIdAndBytes);
        }
    }


    @Override
    public void putTracking(PerceptionResponseOrBuilder tracking) throws InterruptedException
    {
        if (enabled && !getDoStop())
        {
            trackingQueue.put(tracking);
        }
    }


    @Override
    public void putPoses(PerceptionResponse poses) throws InterruptedException
    {
        if (enabled && !getDoStop())
        {
            poseQueue.put(poses);
        }
    }


    @Override
    public void putDistances(PerceptionResponse distances) throws InterruptedException
    {
        if (enabled && !getDoStop())
        {
            distanceQueue.put(distances);
        }
    }


    @Override
    public void putActions(PerceptionResponseOrBuilder actions) throws InterruptedException
    {
        if (enabled && !getDoStop())
        {
            actionsQueue.put(actions);
        }
    }


    @SuppressWarnings("UnnecessaryReturnStatement")
    @Override
    protected void doFirst()
    {
        if (!enabled)
        {
            stop();
            return;
        }

        try
        {
            // Дожидаемся результата, перехватывая возможное исключение
            fpsFuture.get();

            if (resultVideoFile.exists())
            {
                if (!resultVideoFile.delete())
                {
                    LOGGER.error("Can't delete previous result video file '{}' [{}]",
                            resultVideoFile.getAbsolutePath(), managerName);
                    pipelineStopper.run();
                    return;
                }
            }

            if (!resultVideoFile.getParentFile().exists())
            {
                if (!resultVideoFile.getParentFile().mkdirs())
                {
                    LOGGER.error("Can't create parent directories for the resulting video '{}' [{}]",
                            resultVideoFile.getAbsolutePath(), managerName);
                    pipelineStopper.run();
                    return;
                }
            }
        }
        catch (Exception e)
        {
            if (!getDoStop())
            {
                LOGGER.error("Exception occurred during manager initialization [{}]", managerName);
                pipelineStopper.run();
            }
        }
    }


    @Override
    protected void process() throws Exception
    {
        if (!isAllFrameDataAvailable())
        {
            ensureNextFrameData();
        }

        if (currentFrameInfo.frameId == -1)
        {
            submitLastFrameId(-1);
            stop();
            return;
        }

        try
        {
            if (frameRecorder == null)
            {
                frameRecorder = new FFmpegFrameRecorder(resultVideoFile, currentFrameInfo.frameMat.arrayWidth(),
                        currentFrameInfo.frameMat.arrayHeight());
                frameRecorder.setFormat("mp4");
                frameRecorder.setFrameRate(fpsFuture.get());
                frameRecorder.setVideoQuality(0);
                frameRecorder.start();
            }

            Mat frameMat = getFrameWithPerceptionResult();
            Frame f = matConverter.convert(frameMat);
            frameRecorder.record(f);
            frameMat.close();

            submitLastFrameId(currentFrameInfo.frameId);

            clearFrameData();
        }
        catch (Exception e)
        {
            if (!getDoStop())
            {
                LOGGER.error("Exception occurred during writing perception result [{}]", managerName, e);
                pipelineStopper.run();
            }
            throw e;
        }
    }


    @SuppressWarnings("BooleanMethodIsAlwaysInverted")
    private boolean isAllFrameDataAvailable()
    {
        return currentFrameInfo != null && currentTrackingResult != null && currentDistanceResult != null
                && currentActionsResult != null && currentPoseData != null
                && currentFrameInfo.frameId == currentTrackingResult.getFrameId()
                && currentFrameInfo.frameId == currentPoseData.getFrameId()
                && currentFrameInfo.frameId == currentDistanceResult.getFrameId()
                && currentFrameInfo.frameId == currentActionsResult.getFrameId();
    }


    private void ensureNextFrameData() throws InterruptedException
    {
        while (!isAllFrameDataAvailable())
        {
            if (currentFrameInfo == null)
            {
                currentFrameInfo = frameQueue.take();
            }

            if (currentTrackingResult == null)
            {
                currentTrackingResult = trackingQueue.take();
            }

            if (currentPoseData == null)
            {
                currentPoseData = poseQueue.take();
            }

            if (currentDistanceResult == null)
            {
                currentDistanceResult = distanceQueue.take();
            }

            if (currentActionsResult == null)
            {
                currentActionsResult = actionsQueue.take();
            }

            if (currentFrameInfo.frameId != currentTrackingResult.getFrameId())
            {
                if (currentFrameInfo.frameId == -1 || currentTrackingResult.getFrameId() != -1
                        && currentTrackingResult.getFrameId() < currentFrameInfo.frameId)
                {
                    LOGGER.warn("Missing a frame with id {} [{}]", currentTrackingResult.getFrameId(), managerName);
                    currentTrackingResult = trackingQueue.take();
                }
                else
                {
                    LOGGER.warn("Missing a tracking result with frame id {} [{}]", currentFrameInfo.frameId,
                            managerName);
                    currentFrameInfo = frameQueue.take();
                }
            }

            if (currentFrameInfo.frameId != currentDistanceResult.getFrameId())
            {
                if (currentFrameInfo.frameId == -1 || currentDistanceResult.getFrameId() != -1
                        && currentDistanceResult.getFrameId() < currentFrameInfo.frameId)
                {
                    LOGGER.warn("Missing a frame with id {} [{}]", currentDistanceResult.getFrameId(), managerName);
                    currentDistanceResult = distanceQueue.take();
                }
                else
                {
                    LOGGER.warn("Missing a distance result with frame id {} [{}]", currentFrameInfo.frameId,
                            managerName);
                    currentFrameInfo = frameQueue.take();
                }
            }


            if (currentFrameInfo.frameId != currentPoseData.getFrameId())
            {
                if (currentFrameInfo.frameId == -1 || currentPoseData.getFrameId() != -1
                        && currentPoseData.getFrameId() < currentFrameInfo.frameId)
                {
                    LOGGER.warn("Missing a frame with id {} [{}]", currentPoseData.getFrameId(), managerName);
                    currentPoseData = poseQueue.take();
                }
                else
                {
                    LOGGER.warn("Missing a pose result with frame id {} [{}]", currentFrameInfo.frameId,
                            managerName);
                    currentFrameInfo = frameQueue.take();
                }
            }

            if (currentFrameInfo.frameId != currentActionsResult.getFrameId())
            {
                if (currentFrameInfo.frameId == -1 || currentActionsResult.getFrameId() != -1
                        && currentActionsResult.getFrameId() < currentFrameInfo.frameId)
                {
                    LOGGER.warn("Missing a frame with id {} [{}]", currentDistanceResult.getFrameId(), managerName);
                    currentActionsResult = actionsQueue.take();
                }
                else
                {
                    LOGGER.warn("Missing an actions result with frame id {} [{}]", currentFrameInfo.frameId,
                            managerName);
                    currentFrameInfo = frameQueue.take();
                }
            }
        }
    }


    private void clearFrameData()
    {
        currentFrameInfo = null;
        currentTrackingResult = null;
        currentPoseData = null;
        currentDistanceResult = null;
        currentActionsResult = null;
    }


    /**
     * Возвращает текущий кадр, на котором изображены результаты машинного восприятия - обрамляющие рамки объектов,
     * скелеты людей и так далее.
     *
     * @return текущий кадр с результатами машинного восприятия
     */
    private Mat getFrameWithPerceptionResult()
    {
        Map<Long, PoseEstimationEntry> poses = currentPoseData.getPose3D().getEntriesList().stream()
                .collect(Collectors.toMap(PoseEstimationEntry::getTrackId, p -> p));

        Mat img = new Mat(currentFrameInfo.frameMat);
        for (int i = 0; i < currentTrackingResult.getTrackingOrBuilder().getEntriesCount(); i++)
        {
            TrackingEntryOrBuilder te = currentTrackingResult.getTrackingOrBuilder().getEntries(i);
            DetectionEntryOrBuilder de = te.getDetection();
            DistanceEstimationEntry dist = currentDistanceResult.getDistance().getEntries(i);

            Scalar color = COLORS[(int) (te.getId() % COLORS.length)];
            opencv_imgproc.rectangle(img, new Rect(de.getBoxTopLeftX(), de.getBoxTopLeftY(),
                            de.getBoxBottomRightX() - de.getBoxTopLeftX(),
                            de.getBoxBottomRightY() - de.getBoxTopLeftY()),
                    color, 1, opencv_imgproc.LINE_8, 0);
            opencv_imgproc.putText(img, de.getClassId() + " " + te.getId() + " " + dist.getPoints(0).getZ(),
                    new Point(de.getBoxTopLeftX(), Math.max(15, de.getBoxTopLeftY())),
                    opencv_imgproc.FONT_HERSHEY_SIMPLEX, 0.4, color);

            if (de.getClassId() == DetectionClass.PERSON.getId())
            {
                PoseEstimationEntry pe = poses.get(te.getId());
                for (int pInd = 0; pInd < pe.getJointsCount(); pInd++)
                {
                    Point3d p = pe.getJoints(pInd);
                    Point3d parentP = pe.getJoints(POSE_PARENT_IDS[pInd]);

                    opencv_imgproc.line(img, new Point((int) p.getX(), (int) p.getY()),
                            new Point((int) parentP.getX(), (int) parentP.getY()), color, 1,
                            opencv_imgproc.LINE_8, 0);
                }

                for (int pInd = 0; pInd < dist.getPointsCount(); pInd++)
                {
                    Point3d p = pe.getJoints(pInd);
                    opencv_imgproc.circle(img, new Point((int) p.getX(), (int) p.getY()), 1,
                            COLORS[(int) ((te.getId() + 1) % COLORS.length)]);
                }
            }
        }

        Scalar blackColor = new Scalar(0, 0, 0, 1);
        int i = 0;
        for (Action a : currentActionsResult.getActions().getEntriesList())
        {
            ActionClass ac = ActionClass.findById(a.getAction());
            if (ac == null)
            {
                continue;
            }
            opencv_imgproc.putText(img, ac.name() + " " + a.getFirstPersonId() + " "
                            + (a.getSecondPersonId() > 0 ? a.getSecondPersonId() : ""),
                    new Point(10, 15 + i++ * 20),
                    opencv_imgproc.FONT_HERSHEY_SIMPLEX, 0.4, blackColor);
        }

        return img;
    }


    @Override
    protected void doLast()
    {
        // Очищаем для того, чтобы менеджер-источник не сидел на блокировке в случае переотправки завершающего сообщения
        frameQueue.clear();
        trackingQueue.clear();
        poseQueue.clear();
        distanceQueue.clear();
        actionsQueue.clear();

        matConverter.close();

        if (frameRecorder != null)
        {
            try
            {
                frameRecorder.close();
            }
            catch (FrameRecorder.Exception e)
            {
                throw new RuntimeException(e);
            }
        }
    }


    @Override
    public String getName()
    {
        return managerName;
    }


    @Override
    protected boolean isDaemon()
    {
        return false;
    }
}
