package com.devbeh.managers;


import java.time.Duration;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.Semaphore;

import com.devbeh.Actor;
import com.devbeh.FrameIdAndBytes;
import com.devbeh.MetricService;
import com.devbeh.ReducibleSemaphore;
import com.devbeh.data.DetectionEntry;
import com.devbeh.data.PerceptionRequest;
import com.devbeh.data.PerceptionResponse;
import com.devbeh.data.PerceptionResponseOrBuilder;
import com.devbeh.data.TrackInfoEntry;
import com.devbeh.data.TrackerFrameData;
import com.devbeh.data.TrackingEntry;
import com.devbeh.data.TrackingEntryOrBuilder;
import com.devbeh.kafka.KafkaUtils;
import com.devbeh.kafka.TrackerFrameDataDeserializer;
import com.devbeh.kafka.TrackerFrameDataSerializer;
import com.google.protobuf.ByteString;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.clients.consumer.OffsetAndTimestamp;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * Отправляет кадры на обработку сервису отслеживания объектов, ожидает результат обработки, отображает назначенные
 * трекером идентификаторы объектов на глобальные идентификаторы и пересылает результат обработки потребителям.
 * Использование глобальных идентификаторов необходимо, поскольку при перезапуске компонента отслеживания назначенные
 * идентификаторы будут потеряны.
 */
public class TrackingManager extends AbstractDistributor implements FrameConsumer, DetectionConsumer
{
    private static final Logger LOGGER = LoggerFactory.getLogger(TrackingManager.class);

    /**
     * Отображает глобальный идентификатор трека на информацию о треке.
     */
    private final Map<Long, TrackInfo> globalTrackIdToTrackInfo = new HashMap<>();

    /**
     * Отображает идентификатор трека, назначенный трекером, на глобальный идентификатор трека.
     */
    private final Map<Long, Long> trackIdToGlobalTrackId = new HashMap<>();

    /**
     * Отображает идентификатор последнего кадра, на котором присутствовали объекты, на набор глобальных
     * идентификаторов объектов.
     */
    private final LinkedHashMap<Long, Set<Long>> lastFrameIdToGlobalTrackIds = new LinkedHashMap<>();

    // Получатели результатов отслеживания
    private final DistanceEstimationManager dem;
    private final PoseEstimationManager pem;
    private final ActionsManager am;
    private final KBManager kbm;
    private final PerceptionVisualizationManager pvm;

    /**
     * Очередь кадров для обработки.
     */
    private final BlockingQueue<FrameIdAndBytes> frameQueue;

    /**
     * Очередь результатов детектирования.
     */
    private final BlockingQueue<PerceptionResponseOrBuilder> detectionQueue;

    /**
     * Используется для отправки данных для анализа компоненту отслеживания людей и объектов.
     */
    private final TrackingRequestKafkaSender requestKafkaSender;

    /**
     * Имя данного дистрибьютора. Используется при логировании.
     */
    private final String distributorName;

    /**
     * Адрес сервера Kafka.
     */
    private final String bootstrapServer;

    /**
     * Имя пользователя Kafka для SASL аутентификации или пустая строка/null.
     */
    private final String username;

    /**
     * Пароль пользователя Kafka для SASL аутентификации или пустая строка/null.
     */
    private final String password;

    /**
     * Идентификатор клиента (и группы) Kafka, используемый для чтения топика запросов.
     */
    private final String clientId;

    /**
     * Партиция для записи изменения состояния менеджера трекинга. Каждая запись соответствует изменению состояния
     * после обработки соответствующей записи с результатом обработки кадра.
     */
    private final TopicPartition tmStateTopicPartition;

    /**
     * Максимальная длительность перекрытия объекта, после чего объект будет считаться потерянным. Если после этого
     * объект снова появится, он будет считаться новым объектом.
     */
    private final int maxOccludedDurationSec;

    /**
     * Предоставляет fps обрабатываемого видеопотока.
     * Может вызываться только после инициализации начального запрашиваемого кадра, поэтому можно использовать в любом
     * месте {@link #postDoFirst(long)}.
     */
    private final Future<Double> fpsFuture;

    /**
     * Максимальная длительность перекрытия объекта в кадрах. Задается во время инициализации на основе fps.
     */
    private int maxOccludedDurationFrames;

    /**
     * Счет глобальных идентификаторов треков.
     */
    private long globalTrackIdCounter;

    /**
     * Используется для записи изменений состояния в топик.
     */
    private KafkaProducer<?, TrackerFrameData> tmStateProducer;


    public TrackingManager(String videoName, int videoId, String bootstrapServer, String username, String password,
            String requestTopicName, String responseTopicPrefix, String tmStateTopicPrefix, int maxOccludedDurationSec,
            Future<Double> fpsFuture, int maxQueuedRequests, int maxProcessingRequests, DistanceEstimationManager dem,
            PoseEstimationManager pem, ActionsManager am, KBManager kbm, PerceptionVisualizationManager pvm,
            Runnable pipelineStopper, MetricService metricService)
    {
        this(videoName, videoId, bootstrapServer, username, password, "tracking-manager-" + videoId,
                new TopicPartition(requestTopicName, videoId), new TopicPartition(responseTopicPrefix + videoId, 0),
                tmStateTopicPrefix, maxOccludedDurationSec, fpsFuture, new CompletableFuture<>(), maxQueuedRequests,
                maxProcessingRequests, new ReducibleSemaphore(0), dem, pem, am, kbm, pvm, pipelineStopper,
                metricService);
    }


    private TrackingManager(String videoName, int videoId, String bootstrapServer, String username, String password,
            String clientId, TopicPartition requestTopicPartition, TopicPartition responseTopicPartition,
            String tmStateTopicPrefix, int maxOccludedDurationSec, Future<Double> fpsFuture,
            CompletableFuture<Long> initialFrameIdFuture, int maxQueuedRequests, int maxProcessingRequests,
            ReducibleSemaphore processingRequestsSem, DistanceEstimationManager dem, PoseEstimationManager pem,
            ActionsManager am, KBManager kbm, PerceptionVisualizationManager pvm, Runnable pipelineStopper,
            MetricService metricService)
    {
        super("TrackingDistributor-" + videoName, videoName + ".tracking.",
                bootstrapServer, username, password, clientId, requestTopicPartition, responseTopicPartition,
                initialFrameIdFuture, maxProcessingRequests, processingRequestsSem, pipelineStopper, metricService);

        this.distributorName = "TrackingDistributor-" + videoName;
        this.dem = dem;
        this.pem = pem;
        this.am = am;
        this.kbm = kbm;
        this.pvm = pvm;
        this.frameQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.detectionQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.requestKafkaSender = new TrackingRequestKafkaSender("TrackingManager-" + videoName, videoId,
                bootstrapServer, username, password, clientId, requestTopicPartition, initialFrameIdFuture,
                processingRequestsSem);
        this.bootstrapServer = bootstrapServer;
        this.username = username;
        this.password = password;
        this.clientId = clientId;
        this.tmStateTopicPartition = new TopicPartition(tmStateTopicPrefix + videoId, 0);
        this.maxOccludedDurationSec = maxOccludedDurationSec;
        this.fpsFuture = fpsFuture;
    }


    @Override
    protected Actor getRequestKafkaSender()
    {
        return requestKafkaSender;
    }


    @Override
    public void putFrame(FrameIdAndBytes frameIdAndBytes) throws InterruptedException
    {
        if (!getDoStop())
        {
            // Если данный менеджер завершил свою работу, игнорируем дальнейшие сообщения
            frameQueue.put(frameIdAndBytes);
        }
    }


    @Override
    public void putDetection(PerceptionResponseOrBuilder detection) throws InterruptedException
    {
        if (!getDoStop())
        {
            // Если данный менеджер завершил свою работу, игнорируем дальнейшие сообщения
            detectionQueue.put(detection);
        }
    }


    @Override
    protected long getInitialLastDistributedFrameId() throws InterruptedException
    {
        return Math.min(
                Math.min(dem.getInitialLastConsumedFrameId(), pem.getInitialLastConsumedFrameId()),
                Math.min(kbm.getInitialLastConsumedFrameId(), pvm.getInitialLastConsumedFrameId()));
    }


    @Override
    protected void postDoFirst(long lastDistributedFrameId) throws InterruptedException
    {
        double fps;
        try
        {
            fps = fpsFuture.get();
        }
        catch (ExecutionException e)
        {
            // Shouldn't happen!
            handleCriticalError("Exception occurred during fps initialization", e);
            return;
        }

        maxOccludedDurationFrames = (int) (fps * maxOccludedDurationSec);

        loadTmHistory(lastDistributedFrameId);

        if (getDoStop())
        {
            return;
        }

        tmStateProducer = KafkaUtils.buildKafkaProducer(bootstrapServer, username, password, clientId + "-state",
                new TrackerFrameDataSerializer());
    }


    /**
     * Инициализирует {@link #globalTrackIdCounter}, {@link #globalTrackIdToTrackInfo}, {@link #trackIdToGlobalTrackId},
     * {@link #lastFrameIdToGlobalTrackIds}.
     *
     * @param lastDistributedFrameId номер последнего обработанного кадра или -1 в случае, если ни один кадр еще не был
     *                               обработан
     */
    private void loadTmHistory(long lastDistributedFrameId)
    {
        if (lastDistributedFrameId == -1)
        {
            // Загрузка истории не требуется
            return;
        }

        KafkaConsumer<?, TrackerFrameData> tmStateConsumer = KafkaUtils.buildKafkaConsumer(bootstrapServer, username,
                password, clientId + "-state", clientId + "-state", tmStateTopicPartition,
                new TrackerFrameDataDeserializer());

        Collection<TopicPartition> tmStateTPCollection = Collections.singleton(tmStateTopicPartition);
        long beginningOffset = tmStateConsumer.beginningOffsets(tmStateTPCollection).get(tmStateTopicPartition);
        long endOffset = tmStateConsumer.endOffsets(tmStateTPCollection).get(tmStateTopicPartition);

        if (beginningOffset == endOffset)
        {
            LOGGER.error("State history is missing (topic is empty) [{}]. Last distributed frame id: {}",
                    distributorName, lastDistributedFrameId);
            return;
        }

        long initialTmStateFrameId = Math.max(0, lastDistributedFrameId + 1 - maxOccludedDurationFrames);
        OffsetAndTimestamp initialTmStateOffsetAndFrameId = tmStateConsumer.offsetsForTimes(
                Map.of(tmStateTopicPartition, initialTmStateFrameId)).get(tmStateTopicPartition);

        if (initialTmStateOffsetAndFrameId == null)
        {
            // Shouldn't happen!
            LOGGER.error("Required state history records are missing from topic ending [{}]", distributorName);
            // Смещаемся на последнюю запись для инициализации globalFrameIdCounter.
            // Данная запись находится до начала необходимого интервала истории.
            // Гарантируется, что хотя бы одна запись будет присутствовать в топике.
            tmStateConsumer.seek(tmStateTopicPartition, endOffset - 1);
        }
        else
        {
            if (initialTmStateOffsetAndFrameId.timestamp() > lastDistributedFrameId)
            {
                // Shouldn't happen!
                LOGGER.error("Required state history records are missing from topic beginning [{}]", distributorName);
            }
            tmStateConsumer.seek(tmStateTopicPartition, initialTmStateOffsetAndFrameId.offset());
        }

        ArrayList<ConsumerRecord<?, TrackerFrameData>> stateHistory = new ArrayList<>();
        initialTmStateCycle:
        while (!getDoStop())
        {
            for (ConsumerRecord<?, TrackerFrameData> r : tmStateConsumer.poll(Duration.ofMillis(100)))
            {
                globalTrackIdCounter = Math.max(globalTrackIdCounter, r.value().getGlobalTrackIdCounter());

                if (initialTmStateFrameId <= r.timestamp() && r.timestamp() <= lastDistributedFrameId)
                {
                    while (!stateHistory.isEmpty() &&
                            stateHistory.get(stateHistory.size() - 1).timestamp() >= r.timestamp())
                    {
                        stateHistory.remove(stateHistory.size() - 1);
                    }
                    stateHistory.add(r);
                }

                if (r.timestamp() >= lastDistributedFrameId || r.offset() == endOffset - 1)
                {
                    break initialTmStateCycle;
                }
            }
        }

        tmStateConsumer.close();

        if (getDoStop())
        {
            return;
        }

        for (ConsumerRecord<?, TrackerFrameData> r : stateHistory)
        {
            long frameId = r.timestamp();
            for (TrackInfoEntry e : r.value().getEntriesList())
            {
                globalTrackIdToTrackInfo.put(e.getGlobalTrackId(), new TrackInfo(e.getTrackId(), frameId,
                        e.getDetectionId()));
            }
        }

        for (Map.Entry<Long, TrackInfo> ti : globalTrackIdToTrackInfo.entrySet())
        {
            trackIdToGlobalTrackId.put(ti.getValue().trackId, ti.getKey());
            lastFrameIdToGlobalTrackIds.computeIfAbsent(ti.getValue().lastFrameId, f -> new HashSet<>())
                    .add(ti.getKey());
        }
    }


    @Override
    protected void modifyDistributorState(ConsumerRecord<?, PerceptionResponse> record, long lastDistributedFrameId)
            throws InterruptedException
    {
        long frameId = record.value().getFrameId();
        if (frameId == -1 || frameId < lastDistributedFrameId + 1 - maxOccludedDurationFrames)
        {
            return;
        }

        TrackerFrameData.Builder stateChange = TrackerFrameData.newBuilder();

        long detectionIdCounter = record.value().getTrackingOrBuilder().getEntriesOrBuilderList().stream()
                .mapToLong(TrackingEntryOrBuilder::getDetectionInd).max().orElse(0) + 1; // Начиная с 1

        if (lastDistributedFrameId < frameId)
        {
            // Новый обработанный кадр

            for (TrackingEntryOrBuilder te : record.value().getTrackingOrBuilder().getEntriesOrBuilderList())
            {
                long trackId = te.getId();
                long detectionId = te.getDetectionInd() == 0 ? detectionIdCounter++ : te.getDetectionInd();
                Long globalTrackId = trackIdToGlobalTrackId.get(trackId);
                if (globalTrackId == null)
                {
                    // Новый трек
                    globalTrackId = globalTrackIdCounter++;
                    globalTrackIdToTrackInfo.put(globalTrackId, new TrackInfo(trackId, frameId, detectionId));
                    trackIdToGlobalTrackId.put(trackId, globalTrackId);
                }
                else
                {
                    // Существующий трек
                    TrackInfo ti = globalTrackIdToTrackInfo.get(globalTrackId);
                    Set<Long> gIds = lastFrameIdToGlobalTrackIds.get(ti.lastFrameId);
                    gIds.remove(globalTrackId);
                    if (gIds.isEmpty())
                    {
                        lastFrameIdToGlobalTrackIds.remove(ti.lastFrameId);
                    }
                    ti.lastFrameId = frameId;
                }
                lastFrameIdToGlobalTrackIds.computeIfAbsent(frameId, f -> new HashSet<>()).add(globalTrackId);

                // Добавляем запись изменения состояния
                stateChange.addEntriesBuilder().setGlobalTrackId(globalTrackId).setTrackId(trackId)
                        .setDetectionId(detectionId);

                // Удаляем старые треки
                Iterator<Map.Entry<Long, Set<Long>>> lastFrameIdIterator = lastFrameIdToGlobalTrackIds.entrySet()
                        .iterator();
                while (lastFrameIdIterator.hasNext())
                {
                    Map.Entry<Long, Set<Long>> e = lastFrameIdIterator.next();
                    if (e.getKey() >= frameId - maxOccludedDurationFrames)
                    {
                        break;
                    }
                    lastFrameIdIterator.remove();

                    for (long gId : e.getValue())
                    {
                        TrackInfo ti = globalTrackIdToTrackInfo.remove(gId);
                        trackIdToGlobalTrackId.remove(ti.trackId);
                    }
                }
            }
        }
        else
        {
            // Корректировка trackId

            Map<Long, Long> detectionIdToTrackId = new HashMap<>();
            for (TrackingEntryOrBuilder te : record.value().getTrackingOrBuilder().getEntriesOrBuilderList())
            {
                long trackId = te.getId();
                long detectionId = te.getDetectionInd() == 0 ? detectionIdCounter++ : te.getDetectionInd();
                detectionIdToTrackId.put(detectionId, trackId);
            }

            Set<Long> gIds = lastFrameIdToGlobalTrackIds.get(frameId);
            if (gIds != null)
            {
                for (long globalTrackId : gIds)
                {
                    TrackInfo ti = globalTrackIdToTrackInfo.get(globalTrackId);
                    Long newTrackId = detectionIdToTrackId.get(ti.detectionId);
                    if (newTrackId == null || ti.trackId == newTrackId)
                    {
                        // Теоретически, newTrackId может отсутствовать, если трекер вернет новый набор detection
                        // при запуске его с новой конфигурацией. Поэтому минимизируем возможную ошибку
                        continue;
                    }
                    trackIdToGlobalTrackId.remove(ti.trackId);
                    trackIdToGlobalTrackId.put(newTrackId, globalTrackId);
                    ti.trackId = newTrackId;

                    // Добавляем запись изменения состояния
                    stateChange.addEntriesBuilder().setGlobalTrackId(globalTrackId).setTrackId(newTrackId)
                            .setDetectionId(ti.detectionId);
                }
            }
        }
        stateChange.setGlobalTrackIdCounter(globalTrackIdCounter);

        while (!getDoStop())
        {
            try
            {
                tmStateProducer.send(new ProducerRecord<>(tmStateTopicPartition.topic(),
                        tmStateTopicPartition.partition(), frameId, null, stateChange.build())).get();
                break;
            }
            catch (ExecutionException e)
            {
                LOGGER.warn("Exception occurred during sending state change to Kafka [{}]", distributorName);
                sleep(1000);
            }
        }
    }


    @Override
    protected void sendToConsumers(PerceptionResponse response) throws InterruptedException
    {
        // Заменяем trackId на глобальный
        // Дополнительную фильтрацию классов объектов можно не выполнять, так как дополнительно могут быть
        // детектированы только люди
        PerceptionResponseOrBuilder correctedResponse = response;
        if (response.getFrameId() != -1)
        {
            PerceptionResponse.Builder resultBuilder = response.toBuilder();
            for (TrackingEntry.Builder te : resultBuilder.getTrackingBuilder().getEntriesBuilderList())
            {
                te.setId(trackIdToGlobalTrackId.get(te.getId()));
            }
            correctedResponse = resultBuilder;
        }

        sendTrackingIfRequired(dem, correctedResponse);
        sendTrackingIfRequired(pem, correctedResponse);
        sendTrackingIfRequired(am, correctedResponse);
        sendTrackingIfRequired(kbm, correctedResponse);
        sendTrackingIfRequired(pvm, correctedResponse);
    }


    public static void sendTrackingIfRequired(TrackingConsumer trackingConsumer, PerceptionResponseOrBuilder tracking)
            throws InterruptedException
    {
        if (tracking.getFrameId() == -1 || trackingConsumer.getInitialLastConsumedFrameId() < tracking.getFrameId())
        {
            trackingConsumer.putTracking(tracking);
        }
    }


    @Override
    protected void postDoLast()
    {
        // Очищаем для того, чтобы менеджер-источник не сидел на блокировке в случае переотправки завершающего сообщения
        frameQueue.clear();
        detectionQueue.clear();

        if (tmStateProducer != null)
        {
            tmStateProducer.close();
        }
    }


    /**
     * Используется для отправки данных компоненту отслеживания людей и объектов.
     */
    private class TrackingRequestKafkaSender extends AbstractRequestKafkaSender
    {
        /**
         * Имя менеджера, используется при логировании.
         */
        private final String managerName;

        /**
         * Текущий обрабатываемый кадр или null.
         */
        private FrameIdAndBytes currentFrameInfo;

        /**
         * Результат детектирования для текущего обрабатываемого кадра или null.
         */
        private PerceptionResponseOrBuilder currentDetectionResult;


        public TrackingRequestKafkaSender(String managerName, int videoId, String bootstrapServer, String username,
                String password, String clientId, TopicPartition requestTopicPartition,
                Future<Long> initialFrameIdFuture, Semaphore processingRequestsSem)
        {
            super(managerName, videoId, bootstrapServer, username, password, clientId, requestTopicPartition,
                    initialFrameIdFuture, processingRequestsSem, TrackingManager.this::handleCriticalError);
            this.managerName = managerName;
        }


        @Override
        protected boolean isAllFrameDataAvailable()
        {
            return currentFrameInfo != null && currentDetectionResult != null
                    && currentFrameInfo.frameId == currentDetectionResult.getFrameId();
        }


        @SuppressWarnings("DuplicatedCode")
        @Override
        protected void ensureNextFrameData() throws InterruptedException
        {
            while (!isAllFrameDataAvailable())
            {
                if (currentFrameInfo == null)
                {
                    currentFrameInfo = frameQueue.take();
                }

                if (currentDetectionResult == null)
                {
                    currentDetectionResult = detectionQueue.take();
                }

                if (currentFrameInfo.frameId != currentDetectionResult.getFrameId())
                {
                    if (currentFrameInfo.frameId == -1 || currentDetectionResult.getFrameId() != -1
                            && currentDetectionResult.getFrameId() < currentFrameInfo.frameId)
                    {
                        LOGGER.warn("Missing a frame with id {} [{}]", currentDetectionResult.getFrameId(),
                                managerName);
                        currentDetectionResult = detectionQueue.take();
                    }
                    else
                    {
                        LOGGER.warn("Missing a detection with frame id {} [{}]", currentFrameInfo.frameId, managerName);
                        currentFrameInfo = frameQueue.take();
                    }
                }
            }
        }


        @Override
        protected long getCurrentFrameId()
        {
            return currentFrameInfo.frameId;
        }


        @Override
        protected void clearFrameData()
        {
            currentFrameInfo = null;
            currentDetectionResult = null;
        }


        @Override
        protected void buildRequestPart(PerceptionRequest.Builder rb)
        {
            rb.getTrackingBuilder().setImage(ByteString.copyFrom(currentFrameInfo.frame));
            for (DetectionEntry de : currentDetectionResult.getDetection().getEntriesList())
            {
                if (de.getBlurredPatch().isEmpty())
                {
                    rb.getTrackingBuilder().addEntries(de);
                }
                else
                {
                    rb.getTrackingBuilder().addEntries(de.toBuilder().clearBlurredPatch());
                }
            }
        }
    }


    /**
     * Содержит информацию о треке.
     */
    private static class TrackInfo
    {
        /**
         * Текущий номер трека, установленный трекером.
         */
        private long trackId;

        /**
         * Идентификатор последнего кадра, на котором присутствовал объект.
         */
        private long lastFrameId;

        /**
         * Идентификатор объекта в рамках последнего кадра.
         */
        private final long detectionId;


        public TrackInfo(long trackId, long lastFrameId, long detectionId)
        {
            this.trackId = trackId;
            this.lastFrameId = lastFrameId;
            this.detectionId = detectionId;
        }
    }
}
